using mmd2timeline.Player.Model;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;

namespace mmd2timeline.Player.Controllers
{
    /// <summary>
    /// 音频播放控制器
    /// </summary>
    internal class AudioPlayController : StartAndEndTimeController
    {
        /// <summary>
        /// 音频简介控制器
        /// </summary>
        AudioClipController _AudioController;

        /// <summary>
        /// 音源
        /// </summary>
        AudioSource _AudioSource;

        /// <summary>
        /// 音乐选择器
        /// </summary>
        public JSONStorableStringChooser audioChooser;

        /// <summary>
        /// 指示音频播放是否已经暂停
        /// </summary>
        public JSONStorableBool IsPaused;

        /// <summary>
        /// 是否正在加载
        /// </summary>
        public bool IsLoading
        {
            get
            {
                if (AudioController != null)
                {
                    if (AudioController.Inited)
                    {
                        return AudioController.IsLoading;
                    }
                }

                return true;
            }
        }

        /// <summary>
        /// 是否正在等待播放
        /// </summary>
        internal bool IsWaitingPlay = false;

        /// <summary>
        /// 是否有音频剪辑
        /// </summary>
        public bool HasAudio
        {
            get { return AudioController.HasAudio; }
        }

        /// <summary>
        /// 播放中的音频剪辑
        /// </summary>
        public NamedAudioClip PlayingAudio
        {
            get
            {
                return AudioController.PlayingAudio;
            }
        }

        /// <summary>
        /// 是否正在播放
        /// </summary>
        public bool IsPlaying
        {
            get { return _AudioSource.isPlaying; }
        }

        /// <summary>
        /// 设置音量
        /// </summary>
        /// <param name="volume"></param>
        public void SetVolume(float volume)
        {
            this._AudioSource.volume = volume;
        }

        /// <summary>
        /// 获取音频剪辑控制器
        /// </summary>
        private AudioClipController AudioController
        {
            get
            {
                if (_AudioController == null)
                {
                    _AudioController = new AudioClipController();
                }
                return _AudioController;
            }
        }

        /// <summary>
        /// 音频播放控制器的单例
        /// </summary>
        private static AudioPlayController _instance;
        private static object _lock = new object();

        /// <summary>
        /// 音频播放控制器的单例
        /// </summary>
        public static AudioPlayController GetInstance()
        {
            lock (_lock)
            {
                if (_instance == null)
                {
                    _instance = new AudioPlayController();
                }

                return _instance;
            }
        }
        /// <summary>
        /// 音频暂停回调委托
        /// </summary>
        /// <param name="isPaused"></param>
        public delegate void AudioIsPausedCallback(bool isPaused);

        /// <summary>
        /// 当播放暂停变化的事件
        /// </summary>
        public AudioIsPausedCallback OnAudioIsPaused;

        /// <summary>
        /// 音频播放控制器
        /// </summary>
        private AudioPlayController()
        {
            // 初始化
            base.Init("");

            // 初始化音频源
            _AudioSource = URLAudioClipManager.singleton.testAudioSource;

            // 是否已经暂停
            IsPaused = new JSONStorableBool("isPaused", false, new JSONStorableBool.SetBoolCallback((b) => OnAudioIsPaused?.Invoke(b)));

            // 音频选择器
            this.audioChooser = new JSONStorableStringChooser(Lang.Get("Music Path:"), noneStrings, noneStrings, "None", Lang.Get("Music"), new JSONStorableStringChooser.SetStringCallback((p) =>
            {
                if (p == this.audioChooser.defaultVal)
                {
                    this.Stop(1);

                    PlayingAudio?.manager?.RemoveAllClips();
                    AudioController.Clear();
                }
                else
                {
                    this.LoadAudio(p);
                }
            }));
        }

        /// <summary>
        /// 当进度更改后执行的方法
        /// </summary>
        /// <param name="value"></param>
        protected override void OnProgressChange(float value)
        {
            Debug.Log("audio OnProgressChange "+value);
            base.OnProgressChange(value);
            if (value != _AudioSource.time)
            {
                SetAudioTime(value);
            }
        }

        /// <summary>
        /// 设置开始时间
        /// </summary>
        /// <param name="time"></param>
        protected override void SetStartTime(float time)
        {
            base.SetStartTime(time);

            // 设置音频播放进度
            this.SetAudioTime(time);
        }

        /// <summary>
        /// 设置结束时间
        /// </summary>
        /// <param name="time"></param>
        protected override void SetEndTime(float time)
        {
            base.SetEndTime(time);
        }

        /// <summary>
        /// 检查是否已经播放到末尾
        /// </summary>
        /// <param name="value"></param>
        /// <returns></returns>
        protected override bool CheckEnd(float value)
        {
            // 有音频的
            if (this.HasAudio)
            {
                return (base.CheckEnd(value))
                    && !this.IsLoading
                    && !this.IsWaitingPlay;
            }
            else // 无音频的
            {
                return base.CheckEnd(value);
            }
        }

        /// <summary>
        /// 检查是否播放完成
        /// </summary>
        /// <param name="value"></param>
        internal bool CheckPlayEnd(float value)
        {
            if (this.CheckEnd(value))
            {
                this.IsEnd = true;
                OnEnd?.Invoke();
                return true;
            }

            return false;
        }

        /// <summary>
        /// 初始化开始和结束时间
        /// </summary>
        /// <param name="start"></param>
        /// <param name="end"></param>
        internal void InitPlay(List<string> choices, List<string> displayChoices, string choice, float start, float end)
        {
            this.audioChooser.displayChoices = displayChoices;
            this.audioChooser.choices = choices;
            this.audioChooser.val = choice;

            if (choice != noneStrings.First())
            {
                this.IsWaitingPlay = true;
                this.InitStartAndEndTime(start, end);
            }
            else
            {
                this.playProgress.val = 0;
                this.IsEnd = false;

                this.LoadAudio(null);
            }
        }

        /// <summary>
        /// 设置播放速度
        /// </summary>
        /// <param name="speed"></param>
        public void SetPlaySpeed(float speed)
        {
            //_AudioSource.pitch = speed;

            //if (speed == 1f)
            //{
            //    _AudioSource.velocityUpdateMode = AudioVelocityUpdateMode.Auto;
            //}
            //else
            //{
            //    _AudioSource.velocityUpdateMode = AudioVelocityUpdateMode.Fixed;
            //}
        }

        /// <summary>
        /// 根据音频播放进度同步播放状态条
        /// </summary>
        public float Update(float time, float speed = 1f)
        {
            var deltaTime = time * speed;

            // 新的进度
            var newProgress = Progress + deltaTime;

            // 有音频
            if (HasAudio)
            {
                // 正在等待播放
                if (IsWaitingPlay)
                {
                    this.Play();

                    IsWaitingPlay = !this.IsPlaying;
                }
                // 检查如果已经结束，则不再往下进行
                else
                {
                    // 没有结束
                    if (!this.CheckPlayEnd(newProgress))
                    {
                        // 如果正在播放，使用音频时间更新进度条
                        if (this.IsPlaying)
                        {
                            // 如果是正常速度，则取音频的播放时间，否则根据计算时间更新进度
                            if (speed != 1f)
                            {
                                _AudioSource.time = newProgress;
                            }
                            UnityEngine.Debug.Log("audio SetProgress "+ newProgress);
                            this.SetProgress(_AudioSource.time);

                        }
                        else
                        {
                            // 如果新的进度小于音频长度
                            if (newProgress < this.AudioController.Length)
                            {
                                this.Play();
                            }

                            this.SetProgress(newProgress);
                        }
                    }
                }
            }
            else
            {
                // 没有结束
                if (!this.CheckPlayEnd(newProgress))
                {
                    this.SetProgress(newProgress);
                }
            }

            return Progress;
        }

        /// <summary>
        /// 前进
        /// </summary>
        /// <param name="s"></param>
        internal void Forward(float s = 1f)
        {
            this.playProgress.val += s;
        }

        /// <summary>
        /// 后退
        /// </summary>
        /// <param name="s"></param>
        internal void Back(float s = 1f)
        {
            this.playProgress.val -= s;
        }

        /// <summary>
        /// 跳到开头
        /// </summary>
        public void StartAudioOver()
        {
            IsEnd = false;
            SetAudioTime(this.startTime.val);
        }

        /// <summary>
        /// 获取音频的播放时间和进度
        /// </summary>
        /// <returns></returns>
        public float GetAudioTime()
        {
            return _AudioSource.time;
        }

        /// <summary>
        /// 音频停止
        /// </summary>
        public void Stop(int test = 0)
        {
            if (IsPaused.val && audioChooser.val != audioChooser.defaultVal)
            {
                LogUtil.Debug($"Stop::AudioUnPause:{test}");

                this.AudioUnPause();
            }

            if (_AudioSource != null && _AudioSource.isPlaying)
            {
                _AudioSource.Stop();
            }
        }

        /// <summary>
        /// 音频暂停
        /// </summary>
        public void AudioPause()
        {
            if (_AudioSource != null)
            {
                IsPaused.val = true;
                _AudioSource.Pause();
            }
        }
        /// <summary>
        /// 音频取消暂停
        /// </summary>
        public void AudioUnPause()
        {
            if (_AudioSource != null)
            {
                _AudioSource.UnPause();
                IsPaused.val = false;
            }
        }

        /// <summary>
        /// 设置音频源的播放时间
        /// </summary>
        /// <param name="time"></param>
        public void SetAudioTime(float time)
        {
            if (time < startTime.val) { time = startTime.val; }

            if (_AudioSource != null)
            {
                _AudioSource.time = time;
            }

            if (time != this.Progress)
            {
                this.SetProgress(time);
            }
        }

        /// <summary>
        /// 播放音频
        /// </summary>
        /// <param name="time"></param>
        /// <returns></returns>
        public float Play()
        {
            if (!_AudioSource.isPlaying && PlayingAudio != null)
            {
                if (PlayingAudio?.clipToPlay?.loadState == AudioDataLoadState.Loaded)
                {
                    _AudioSource.clip = PlayingAudio.sourceClip;
                    _AudioSource.Play();
                }
            }

            return Progress;
        }

        /// <summary>
        /// 检查音频的加载状态
        /// </summary>
        internal void CheckLoad()
        {
            if (!HasAudio)
                return;

            var length = AudioController.TryGetAudioLength();

            if (length.HasValue)
            {
                this.MaxTime = length.Value;
                this.SetProgress(this.startTime.val);

                this.IsWaitingPlay = (this.MaxTime > 0);
                this.OnLengthGot(MaxTime);
            }
            else
            {
                this.IsWaitingPlay = true;
            }
        }

        /// <summary>
        /// 加载指定地址的音频
        /// </summary>
        /// <param name="audioPath"></param>
        private void LoadAudio(string audioPath)
        {
            var pausedBefore = this.IsPaused.val;
            this.Clear(false);
            this.IsEnd = false;
            this.IsPaused.val = pausedBefore;

            AudioController.LoadAudio(audioPath);
        }

        /// <summary>
        /// 清空播放内容
        /// </summary>
        public void Clear(bool cleanChooser = true)
        {
            this.Stop(2);

            AudioController.Clear();

            if (cleanChooser)
            {
                this.MaxTime = 0;
                this.audioChooser.choices = noneStrings;
                this.audioChooser.displayChoices = noneStrings;
                this.audioChooser.val = noneStrings.FirstOrDefault();
            }

            URLAudioClipManager.singleton.RemoveAllClips();
            URLAudioClipManager.singleton.RestoreAllFromDefaults();
        }

        /// <summary>
        /// 销毁时执行的函数
        /// </summary>
        public override void OnDestroy()
        {
            this.Clear();

            this.OnAudioIsPaused = null;
            this.OnProgressChanged = null;
            this.OnEnd = null;
            this._AudioSource = null;

            _instance = null;
            base.OnDestroy();
        }

        /// <summary>
        /// 禁用时执行的函数
        /// </summary>
        public override void OnDisable()
        {
            this.Stop(3);
        }

        /// <summary>
        /// 启用时执行的函数
        /// </summary>
        public override void OnEnable()
        {

        }
    }
}
