﻿using MeshVR;
using MVR.FileManagementSecure;
using SimpleJSON;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using UnityEngine;

namespace mmd2timeline.Player.Model
{
    /// <summary>
    /// 语言处理类
    /// </summary>
    internal class Lang : MSJSONClass
    {
        /// <summary>
        /// 获取指定字符串的翻译
        /// </summary>
        /// <param name="key"></param>
        /// <returns></returns>
        public static string Get(string key)
        {
            return GetInstance().GetValue(key);
        }

        /// <summary>
        /// 获取指定翻译的原始字符串
        /// </summary>
        /// <param name="value"></param>
        /// <returns></returns>
        public static string From(string value)
        {
            return GetInstance().FromKey(value);
        }

        /// <summary>
        /// 获取语言参数
        /// </summary>
        /// <param name="key"></param>
        /// <returns></returns>
        public string GetValue(string key)
        {
            if (this.HasKey(key))
            {
                return this[key];
            }
            else
            {
                this[key] = key;
#if DEBUG
                this.Save(LangFilePath);
#endif

                return key;
            }
        }

        /// <summary>
        /// 获取原始语言
        /// </summary>
        /// <param name="value"></param>
        /// <returns></returns>
        public string FromKey(string value)
        {
            foreach (var key in this.Keys)
            {
                var v = this[key].Value;

                if (value.Contains(v))
                {
                    return key;
                }
            }

            return value;
        }

        /// <summary>
        /// 语言处理对象
        /// </summary>
        private Lang()
        {

        }

        /// <summary>
        /// 生成语言配置文件
        /// </summary>
        internal void GenerateProfile()
        {
            this.Save(LangFilePath);
        }

        /// <summary>
        /// 加载语言配置文件
        /// </summary>
        internal void LoadProfile()
        {
            // 如果文件存在，则加载语言配置文件
            if (FileManagerSecure.FileExists(LangFilePath))
            {
                this.Load(LangFilePath);
            }
        }

        /// <summary>
        /// 根据系统语言加载语言字典
        /// </summary>
        private void LoadByLanguage()
        {
            LogUtil.Debug($"{Application.systemLanguage}");

            switch (Application.systemLanguage)
            {
                case SystemLanguage.Chinese:
                case SystemLanguage.ChineseSimplified:
                case SystemLanguage.ChineseTraditional:
                    LoadChinese();
                    break;
                default:
                    break;
            }
        }

        private static readonly string LangFilePath = "Saves\\PluginData\\MMD2Timeline\\lang.json";
        private static Lang _instance;
        private static object _lock = new object();

        /// <summary>
        /// 语言处理类的单例
        /// </summary>
        public static Lang GetInstance()
        {
            lock (_lock)
            {
                if (_instance == null)
                {
                    _instance = new Lang();

                    _instance.LoadByLanguage();

                    _instance.LoadProfile();
                }

                return _instance;
            }
        }

        protected override void BeforeToString()
        {

        }

        /// <summary>
        /// 加载中文
        /// </summary>
        void LoadChinese()
        {
            this["is Ready.\nPlease select the Content to play in the Player."] = "准备完毕\n请在Player中选择要播放的内容。";
            this["<color=#606060><size=40><b>Capturer</b></size>\nBased on MacGruber's SuperShot v13\nSave sequence screenshots for generating video files.</color>"] = "<color=#606060><size=40><b>Capturer</b></size>\n基于 MacGruber's SuperShot v13\n保存序列截图以生成视频文件。</color>";
            this["Depending on the settings, this plugin can be <b>extremely demanding</b> in terms of graphics memory! Reduce the resolution multiplier and/or MSAA when you notice trouble like incomplete images or low performance. \n\nMemory is only allocated for the moment when taking a screenshot and released immediately afterwards."] = "根据不同设置，此插件对显存的要求可能 <b>极高</b>！当发现图像不完整或性能低下等问题时，请降低分辨率乘数和/或 多重采样抗锯齿数值。 \n\n内存只在截图时分配，截图后立即释放。";
            this["Capture Records"] = "采集记录";
            this["Capture Record Info"] = "采集记录信息";
            this["Begin Frame"] = "起始帧";
            this["Start Capture"] = "开始采集";
            this["Recapture"] = "重新采集";
            this["Continue Capture"] = "继续采集";
            this["Render FPS"] = "渲染FPS";
            this["Aspect Ratio"] = "宽高比";
            this["Resolution Output"] = "输出分辨率";
            this["Portait Mode (Flip AspectRatio)"] = "竖屏模式（翻转纵横比）";
            this["Resolution Multiplier"] = "分辨率倍数";
            this["Downscale Method"] = "缩小比例方法";
            this["MSAA Override"] = "多重采样抗锯齿覆盖";
            this["Image Format"] = "图像格式";
            this["<b>General Options</b>"] = "<b>常规选项</b>";
            this["Preview UI scale"] = "预览 UI 比例";
            this["Lock Z Rotation"] = "锁定 Z 轴旋转";
            this["Flashlight"] = "闪光灯";
            this["Setting"] = "设置";
            this["Add New"] = "添加新的";
            this["Remove Current"] = "移除当前的";
            this["Set To Default"] = "设为默认";
            this["Person"] = "人物";
            this["Add Person"] = "添加人物";
            this["Remove Person"] = "移除人物";
            this["Play"] = "播放";
            this["Pause"] = "暂停";
            this["Prev"] = "上一个";
            this["Next"] = "下一个";
            this["Remove from Playlist"] = "从播放列表移除";
            this["Save Settings"] = "保存设置";
            this["Duplicate Files Check"] = "重复文件检查";
            this["Auto Check"] = "自动检查";
            this["Stop Playing When Found"] = "发现重复文件停止播放";
            this["Checking Progress"] = "检查进度";
            this["Camera Motion"] = "镜头动作";
            this["Progress"] = "进度";
            this["Delay"] = "延迟";
            this["Sample Mode"] = "采样模式";
            this["EveryFrame"] = "所有帧";
            this["KeyFrame"] = "关键帧";
            this["Camera Scale"] = "镜头缩放";
            this["Enabled Position Control"] = "开启位置控制";
            this["Enabled Rotation Control"] = "开启方向控制";
            this["Focus On ..."] = "聚焦...";
            this["Position X"] = "位置 X";
            this["Position Y"] = "位置 Y";
            this["Position Z"] = "位置 Z";
            this["Rotation X"] = "方向 X";
            this["Rotation Y"] = "方向 Y";
            this["Rotation Z"] = "方向 Z";
            this["Target Atom"] = "目标原子";
            this["Target Receiver"] = "目标位置";
            this["Start Time"] = "开始时间";
            this["End Time"] = "结束时间";
            this["Music Path:"] = "音频路径";
            this["Music"] = "音频";
            this["Load Playlist"] = "加载播放列表";
            this["Import From Folder"] = "从文件夹载入";
            this["Favorite"] = "收藏";
            this["UnFavorite"] = "取消收藏";
            this["Load Folder"] = "加载文件夹";
            this["Load File"] = "加载文件";
            this["Play Mode"] = "播放模式";
            this["Playlist"] = "播放列表";
            this["Load Favorite"] = "加载收藏";
            this["Load All"] = "加载所有";
            this["Clear All"] = "清空所有";
            this["Export Playlist"] = "导出播放列表";
            this["<color=#000><size=24>Click the buttons above to prepare to start your Dance!</size></color>\n\n<b>Load Playlist</b> - Import the playlist settings from a saved file.\n\n<b>Import From Folder</b> - Load multiple MMDs to the playlist. You should select the upper directory where the MMD files are stored. \n\n<b>Load Folder</b> - Load an MMD into temporary memory and You can test or adjust its play settings.If it works well, you can add it to the playlist.\n\n<b>Load File(vmd/wav/mp3/ogg)</b> - Load a file to the current MMD.\n\n<b>Load Favorites</b> - Load your favorite MMDs to the current Playlist.\n\n<b>Load All</b> - Load All MMDs to the current Playlist."]
                = "<color=#000><size=24>点击上面的按钮，准备开始表演！</size></color>\n\n<b>加载播放列表</b> - 从保存的文件中导入播放列表设置。\n\n<b>从文件夹载入</b> - 将多个 MMD 载入播放列表。请选择存储 MMD 文件的上层目录。 \n\n<b>加载文件夹</b> - 将 MMD 文件夹载入临时存储器，您可以测试或调整其播放设置，设置好后可以将其添加到播放列表中。\n\n<b>加载文件</b> - 加载一个文件到当前 MMD。\n\n<b>加载收藏</b> - 将您收藏的 MMD 载入到当前播放列表。\n\n<b>加载所有</b> - 将所有 MMD 载入当前播放列表。";
            this["Show Camera Settings"] = "显示镜头设置";
            this["Play Speed"] = "播放速度";
            this["\n\nhttps://space.bilibili.com/1016760606\n\n<b>Follow the page get latest versions.</b>\n\n<size=24><b>IMPORTANT!!!</b></size>\n<size=24>The MMD Folder Must In VAM Path.If Your MMD Folder In Other Path, Use The CMD Command 'mklink' Link it to VAM Path.</size>\n"]
                = "\n\nhttps://space.bilibili.com/1016760606\n\n<b>关注以上页面获取最新版本</b>\n\n<size=24><b>重要！！！</b></size>\n<size=24>MMD 文件夹必须位于 VAM 路径下。如果您的 MMD 文件夹位于其他路径下，请使用 CMD 命令 \"mklink \"将其链接到 VAM 目录中。</size>\n";
            this["Show Debug Info"] = "显示调试信息";
            this["Motion Scale"] = "动作缩放";
            this["Sample Speed"] = "采样速度";
            this["Use All Joints Settings"] = "开启所有关节设置";
            this["All Joints Spring Percent"] = "所有关节弹簧倍数";
            this["All Joints Damper Percent"] = "所有关节阻尼倍数";
            this["All Joints Max Velocity"] = "所有关节最大速度倍数";
            this["Enable Face"] = "启用表情";
            this["Show Motion Settings"] = "显示动作设置";
            this["General Settings"] = "通用设置";
            this["Play Volume"] = "播放音量";
            this["Lock Person Position"] = "锁定人物位置";
            this["Motion Position State"] = "动作关节位置设定";
            this["Motion Rotation State"] = "动作关节方向设定";
            this["Physics Mesh Settings"] = "柔性物理网格设置";
            this["Enable Mouth Physics Mesh"] = "启用嘴部柔性物理";
            this["Enable Breast Physics Mesh"] = "启用胸部柔性物理";
            this["Enable Lower Physics Mesh"] = "启用下身柔性物理";
            this["Auto-Correct Height Settings"] = "自动修正高度设置";
            this["Auto-Correct Mode"] = "修正模式";
            this["Fix Height"] = "修正高度";
            this["Floor Height"] = "地板高度";
            this["Enable Foot Free"] = "启用脚部放松";
            this["Enable Foot Off"] = "允许脚部关闭控制";
            this["Enable Knee Off"] = "允许膝盖关闭控制";
            this["Toe Off Height"] = "脚趾关闭高度";
            this["Foot Fix Height"] = "脚部修正高度";
            this["Foot Off Height"] = "脚部关闭高度";
            this["Free Foot Joint Drive X Target"] = "放松脚部关节驱动 X 目标";
            this["Free Foot Hold Rotation Max Force"] = "放松脚部保持旋转最大力";
            this["Free Knee Hold Rotation Max Force"] = "放松膝盖保持旋转最大力";
            this["Free Thigh Hold Rotation Max Force"] = "放松大腿保持旋转最大力";
            this["Camera Settings"] = "镜头设置";
            this["Camera Enabled Non-VR"] = "开启镜头（非VR）";
            this["Camera Enabled in VR"] = "开启镜头（VR）";
            this["Camera Position Smoothing"] = "镜头位置平滑";
            this["Camera Rotation Smoothing"] = "镜头角度平滑";
            this["Deactive when MainHUD Opened"] = "主界面打开时停止镜头";
            this["Only Keyframes"] = "镜头播放只使用关键帧";
            this["Play in WindowCamera"] = "在WindowCamera中播放";
            this["<color=#FF0000><b>Plugin VRAM usage: ~"] = "<color=#FF0000><b>插件 VRAM 使用量: ~";
            this[" GB.</b></color>"] = " GB.</b></color>";
            this["Importing"] = "导入中";
            this["Position"] = "位置";
            this["Close Lower Bones"] = "关闭下身骨骼";
            this["File Type"] = "文件类型";
            this["Length"] = "长度";
            this["Delete"] = "删除";
            this["Remove"] = "移除";
            this["Generate Language Profile"] = "生成语言配置文件";
            this["If you need to set your own language, press the \"Generate Language Profile\" button, it will generate the Saves\\PluginData\\MMD2Timeline\\lang.json file, you can modify the content in it to get your own language profiles."]
                = "如果你需要自己来设定语言，按下“生成语言配置文件”按钮，将会生成Saves\\PluginData\\MMD2Timeline\\lang.json文件，修改里头的内容来获得你自己的语言配置文件。";
            this["Exit Edit Mode"] = "退出编辑模式";
            this["Global Motion Scale"] = "全局动作缩放";
            this["Enable High Heel"] = "启用高跟";
            this["Foot Hold Rotation Max Force"] = "脚部保持最大力";
            this["Foot Joint Drive X Angle"] = "脚部关节X角度";
            this["Toe Joint Drive X Angle"] = "脚趾关节X角度";
            this["Reset Person Physical When Motion Reset"] = "动作重置时重设人物物理";
            this["Reset All Settings To Default"] = "重置所有设置到默认值";
            this["Press ESC to Stop Capture."] = "按ESC键停止采集";
            this["Duplicate Files"] = "重复文件";
            this["Select the files you want to keep and click \"Delete Others\" to add the remaining files to PrepareToDeleteFile (You can run "]
                = "选中你要保存的文件，点击 \"删除其他\" 按钮将其余文件添加到待删除文件清单（你可以运行 ";
            this[" to remove them from the disk.) or click \"Remove Others\" to remove them from the settings."] = " 来把它们从硬盘删除。）或点击 \"移除其他\" 按钮来把他们从设置中移除。";
            this["Delete Others"] = "删除其他";
            this["Remove Others"] = "移除其他";
            this["Capture Completed."] = "录制完成。";
            this["Execute"] = "运行";
            this["in"] = "在";
            this["to create your video file"] = "来创建视频文件。";
            this["Ready For Capture.."] = "准备录制..";
            this["Capturing..."] = "录制中...";
            this["Capture Progress:"] = "录制进度：";
            this["Motion Mode"] = "动作模式";
            this["Custom"] = "自定义";
            this["Fatigue"] = "乏力";
            this["Weak"] = "柔弱";
            this["Smooth"] = "轻柔";
            this["Normal"] = "一般";
            this["Strong"] = "强壮";
            this["Dexterity"] = "灵巧";
            this["Agile"] = "敏捷";
            this["On"] = "开";
            this["Off"] = "关";
            this["Following"] = "跟随";
            this["Hold"] = "保持";
            this["Lock"] = "锁定";
            this["ParentLink"] = "父连接";
            this["PhysicsLink"] = "物理连接";
            this["Comply"] = "遵守";
            this["LookAt"] = "看向";
            this["None"] = "无";
            this["PartOnly"] = "仅超范围部位";
            this["WholeBody"] = "全身";
            this["Default"] = "默认";
            this["Random"] = "随机";
            this["Repeat"] = "重复";
            this["Music"] = "音频";
            this["Camera"] = "镜头";
            this["Motion"] = "动作";
            this["Expression"] = "表情";
            this["VMD"] = "VMD";
            this["Other"] = "其他";
            this["Reset Motion Model\n(Fix motion problems with bone changes)"] = "重置动作模型\n（修复因骨骼变动引起的动作异常）";
            this["The VMD base processing mechanism of this plugin is based on the technical achievements of mmd2timeline, We are looking into the possibility of merging the two plugins, stay tuned."] = "本插件的vmd基础处理机制基于mmd2timeline的技术成果，我们正在研究两个插件合并的可能，敬请期待。";
            this["Import Completed"] = "导入完成";
            this["MMDs Imported."] = "MMD导入完成";
            this["Press OK button to Start the Show."] = "按下OK按钮开始表演";
            this["OK"] = "OK";
            this["Language Settings"] = "语言设置";
            this["Enable Space to Toggle Play/Pause"] = "启用<空格键>切换播放/暂停";
            this["Enable Right Arrow to Forward 1s"] = "启用<右箭头>前进1秒";
            this["Enable Left Arrow to Back 1s"] = "启用<左箭头>后退1秒";
            this["Enable Up Arrow to Play Previous"] = "启用<上箭头>切换上一首";
            this["Enable Down Arrow to Play Next"] = "启用<下箭头>切换下一首";
            this["U - Toggle UI Hidden/Visible"] = "U键 - 切换UI显示/隐藏";
            this["M - Toggle WindowCamera View"] = "M键 - 切换窗口相机视图";
            this["Shortcut Keys"] = "快捷键";
            this["Technical References Catalog:"] = "技术引用清单：";
            //this["MMD2Timeline is based on the LoadScene module in the open source project varManager on github."]
            //    = "MMD2Timeline是在Github上开源项目varManager的LoadScene模块基础上构建的。";
            this["MMD2Timeline's vmd reading, and core functions including but not limited to basic skeleton and motion management are based on mmd2timeline's technical achievements, and we have been explicitly authorized to develop subsequent versions based on the corresponding technical foundation."]
                = "MMD2Timeline的VMD读取，以及包括但不限于基本骨架和动作管理在内的核心运作功能是基于mmd2timeline的技术成果进行开发的，我们已获得明确授权，可在相应技术基础上开发后续版本。";
            this["mmd2timeline has been improved based on Libmmd-for-Unity, and so MMD2Timeline has inherited the technical relationship."]
                = "mmd2timeline基于Libmmd-for-Unity进行了改进，因此MMD2Timeline也继承了这种技术关系。";
            this["MMD2Timeline's audio load Method from the open source project VamFreeMMD."]
                = "MMD2Timeline的音频加载方法来自开源项目VamFreeMMD。";
            this["Some of MMD2Timeline's UI processing uses MacGruber_Utils from MacGruber Essentials, and we've made a few feature tweaks and added some functionality. " +
                "The Capturer module, which we haven't publicly released yet, uses the SuperShot module from that plugin, and we modified its operation rules " +
                "and added memory release handling to meet the needs of frame-by-frame recording.That Plugin available for free under \"CC BY-SA\" license."]
                = "MMD2Timeline 的部分用户界面处理使用了 MacGruber Essentials 中的 MacGruber_Utils，我们还对一些特性进行了调整，并添加了一些功能。" +
                "我们尚未公开发布的Capturer模块使用了该插件中的SuperShot模块，我们修改了它的操作规则，并添加了内存释放处理。该插件在\"CC BY-SA\"许可下免费提供。";
            this["MMD2Timeline's camera focus feature is based on Searchlight improvements.Searchlight basically \"CC BY-NC\"."]
                = "MMD2Timeline的相机对焦功能是基于Searchlight改进而来的，Searchlight基于是\"CC BY-NC\"授权可免费使用。";
            this["MMD2Timeline's follow-the-camera feature is based on Passenger, Passenger complies with the MIT license."]
                = "MMD2Timeline的镜头跟随功能基于Passenger的运行机制进行开发，Passenger遵循MIT许可。";
            this["MMD2Timeline is a free VAM plugin. \nMMD2Timeline code is licensed under the AGPL 3.0 license, except for the code referenced above, which is licensed under the original license. \nDo not use it for unauthorized commercial use or in closed source projects."]
                = "MMD2Timeline是一款免费的VAM插件。\n除上面提到的这部分代码遵循其原始许可外，MMD2Timeline源代码采用AGPL 3.0许可。\n请勿将其用于未经授权的商用或闭源项目中。";
        }
    }
}
