using MacGruber;
using MeshVR;
using mmd2timeline.Player.Controllers;
using mmd2timeline.Player.Model;
using SimpleJSON;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;

namespace mmd2timeline.Player
{
    internal partial class Player : Settings, IMMDController
    {
        #region 播放状态参数

        // 指示是否在播放中
        internal bool IsPlaying
        {
            get
            {
                return _IsPlaying.val;
            }
        }

        /// <summary>
        /// 是否正在播放
        /// </summary>
        JSONStorableBool _IsPlaying;
        #endregion

        /// <summary>
        /// 音频播放控制器
        /// </summary>
        AudioPlayController _AudioPlayController;

        /// <summary>
        /// 状态检查时间。
        /// </summary>
        float _CheckTime = 0f;

        /// <summary>
        /// 检查动画暂停时间
        /// </summary>
        float _CheckFreezeTime = 0f;

        /// <summary>
        /// 最大状态检查时间。超过此时间没有响应，则跳到下一个
        /// </summary>
        const float MAX_CHECK_TIME = 1f;

        /// <summary>
        /// 指示动画是否正在加载中
        /// </summary>
        bool _IsLoading = false;
        /// <summary>
        /// 当前播放的MMD动画
        /// </summary>
        private MMDEntity _CurrentItem = null;

        /// <summary>
        /// 声明存储MMD人物的变量
        /// </summary>
        Dictionary<int, PersonMotionController> _PersonMotionControllers = new Dictionary<int, PersonMotionController>();
        /// <summary>
        /// 镜头动作控制器
        /// </summary>
        CameraMotionController _CameraMotionController = null;

        /// <summary>
        /// MMD文件管理员
        /// </summary>
        MMDFolderController _MMDFolderController = null;

        /// <summary>
        /// 创建人物原子列表变量
        /// </summary>
        List<Atom> _PersonAtoms;

        /// <summary>
        /// 播放切换的按键
        /// </summary>
        KeyCode _PlayToggleKey = KeyCode.Space;

        /// <summary>
        /// 插件目录
        /// </summary>
        string _PluginPath;

        /// <summary>
        /// 指示播放器是否已经锁定
        /// </summary>
        internal bool IsLocked = false;

        public override bool ShouldIgnore()
        {
            return false;
        }

        /// <summary>
        /// 设置对外开放的方法
        /// </summary>
        void SetOpenActions()
        {
            RegisterAction(new JSONStorableAction("Play/Stop", () => this.TogglePlaying()));
            RegisterAction(new JSONStorableAction("Play", () => this.StartPlaying()));
            RegisterAction(new JSONStorableAction("Stop", () => this.StopPlaying()));
            RegisterAction(new JSONStorableAction("Prev", () => this.Prev()));
            RegisterAction(new JSONStorableAction("Next", () => this.Next()));
            RegisterAction(new JSONStorableAction("Import From Folder", () => this._MMDFolderController.ImportFromFolder()));
            RegisterAction(new JSONStorableAction("Load From Folder", () => this._MMDFolderController.LoadFolder()));
            RegisterAction(new JSONStorableAction("Load File", () => this._MMDFolderController.LoadFile()));

            RegisterAction(new JSONStorableAction("Clear All", () => StopPlayAndRun(() => this.StartCoroutine(this.Playlist.ClearList()))));
            RegisterAction(new JSONStorableAction("Remove Current", () => this.Playlist.RemovePlayItem()));

            RegisterAction(new JSONStorableAction("Import From File", () => this.Playlist.BeginImport()));
            RegisterAction(new JSONStorableAction("Export Playlist", () => this.Playlist.BeginExport()));
        }

        /// <summary>
        /// 原子被添加的接收方法
        /// </summary>
        /// <param name="atom"></param>
        void CheckAtomAdded(Atom atom)
        {
            // 实例化MMD人物
            InitPersonMotionController(atom);
        }
        /// <summary>
        /// 原子被移除的接收方法
        /// </summary>
        /// <param name="atom"></param>
        void CheckAtomRemoved(Atom atom)
        {
            // 移除人物
            RemovePersonMotionController(atom);
        }

        /// <summary>
        /// 初始化
        /// </summary>
        public override void Init()
        {
            base.Init();

            InitStatusUI();

            _PluginPath = Utils.GetPluginPath(this);

            _IsPlaying = new JSONStorableBool("isPlaying", false, new JSONStorableBool.SetBoolCallback((p) =>
            {
                if (p)
                {
                    // 隐藏主HUD
                    SuperController.singleton.HideMainHUD();
                }
                SetPlayButton();
            }));

            SetOpenActions();

            PersonAtoms.Clear();

            SuperController.singleton.onAtomAddedHandlers = CheckAtomAdded;
            SuperController.singleton.onAtomRemovedHandlers = CheckAtomRemoved;

            // 初始化镜头
            _CameraMotionController = CameraMotionController.GetInstance();
            _CameraMotionController.player = this;
            _CameraMotionController.OnLoaded += OnCameraLoaded;

            //--音频播放控制器--------------------------------------------------------------
            _AudioPlayController = AudioPlayController.GetInstance();
            _AudioPlayController.OnAudioIsPaused = new AudioPlayController.AudioIsPausedCallback(v =>
            {
                SetPlayButton();
            });

            _AudioPlayController.OnProgressChanged = SyncProgress;

            _AudioPlayController.OnEnd = Next;
            _AudioPlayController.OnLoaded += OnAudioLoaded;

            //--MMD文件管理器--------------------------------------------------------------
            _MMDFolderController = MMDFolderController.GetInstance();
            _MMDFolderController.player = this;
            _MMDFolderController.OnMMDSelected = (e, needAddToList) =>
            {
                if (needAddToList)
                {
                    StartCoroutine(SaveSettings(e, needAddToList));
                }
                else
                {
                    if (e != null)
                    {
                        InEditMode();

                        LoadMMD(e, true);
                    }
                    else
                    {
                        LoadMMD(this.Playlist.CurrentMMD);
                    }
                }
            };
            _MMDFolderController.OnMMDImported += OnMMDImported;

            InitSettingUI();
        }

        /// <summary>
        /// 显示导入进度
        /// </summary>
        /// <param name="info"></param>
        /// <param name="step"></param>
        /// <param name="total"></param>
        void ShowImportProgress(string info, int step, int total)
        {
            if (step < total)
            {
                var msg = "\n" +
                $"<size=35><b>{Lang.Get("Importing")}... {step + 1}/{total}</b></size>" +
                $"\n" +
                $"{info}" +
                $"\n";

                this.ShowTip(message: msg, step: step, total: total);
            }
            else
            {
                var msg = "\n" +
                $"<size=35><b>{Lang.Get("Import Completed")}</b></size>" +
                $"\n" +
                $"\n" +
                $"{info}" +
                $"\n" +
                $"\n" +
                $"{total} {Lang.Get("MMDs Imported.")}" +
                $"\n" +
                $"\n" +
                $"{Lang.Get("Press OK button to Start the Show.")}" +
                $"\n" +
                $"\n";

                this.ShowTip(message: msg, step: step, total: total, action: () =>
                {
                    this.ShowPlayUI();

                    this.Playlist.First();
                });
            }
        }

        /// <summary>
        /// 当MMD导入时执行的方法
        /// </summary>
        /// <param name="path"></param>
        /// <param name="step"></param>
        /// <param name="total"></param>
        void OnMMDImported(MMDEntity entity, string info, int step, int total)
        {
            LogUtil.Debug($"Player::OnMMDImported:{step}/{total}");

            if (entity != null)
            {
                this.Playlist.AddPlayItem(entity, notify: false);
            }

            ShowImportProgress(info, step, total);

            if (step == total)
            {
                // 导入完成后，清理播放列表
                this.Playlist.ClearPlayList();

                this.OnPlaylistUpdated(this.Playlist, this.Playlist.GetPlayList());
            }
        }

        protected new void InitSettingUI()
        {
            // 创建UI
            CreateUI();

            foreach (var atom in SuperController.singleton.GetAtoms())
            {
                CheckAtomAdded(atom);
            }
        }

        #region Save/Load
        public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
        {
            var json = base.GetJSON(includePhysical, includeAppearance, forceStore);
            try
            {
                needsStore = true;
            }
            catch (Exception exc)
            {
                LogUtil.LogError(exc);
            }
            return json;
        }

        public override void RestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, JSONArray presetAtoms = null, bool setMissingToDefault = true)
        {
            base.RestoreFromJSON(jc, restorePhysical, restoreAppearance, presetAtoms, setMissingToDefault);

            try
            {
            }
            catch (Exception exc)
            {
                LogUtil.LogError(exc);
            }
        }

        public override void PostRestore(bool restorePhysical, bool restoreAppearance)
        {
            base.PostRestore(restorePhysical, restoreAppearance);
        }

        #endregion

        /// <summary>
        /// 同步进度条
        /// </summary>
        /// <param name="audioProgress"></param>
        void SyncProgress(float audioProgress)
        {
            Debug.Log("player SyncProgress " + audioProgress);
            var cameraProgress = audioProgress - _AudioPlayController.startTime.val;

            foreach (var item in _PersonMotionControllers)
            {
                var personMotionController = item.Value;

                var motionProgress = audioProgress - _AudioPlayController.startTime.val;

                // 同步人物动作播放进度
                StartCoroutine(SetMMDProgress(personMotionController, motionProgress));
                //SetMMDProgress(personMotionController, motionProgress);
            }

            SetMMDCameraProgress(cameraProgress);

            ProgressUpdated?.Invoke(this._AudioPlayController.Progress);
        }

        internal event Action<float> ProgressUpdated;

        /// <summary>
        /// 音频加载完毕调用的函数
        /// </summary>
        /// <param name="maxtime"></param>
        void OnAudioLoaded(ProgressController sender, float maxtime)
        {
            LogUtil.Debug($"Player::OnAudioLoaded:{maxtime}");

            UpdateMaxMMDLenght();
        }

        /// <summary>
        /// 镜头加载完毕调用的函数
        /// </summary>
        /// <param name="maxtime"></param>
        void OnCameraLoaded(ProgressController sender, float maxtime)
        {
            LogUtil.Debug($"Player::OnCameraLoaded:{maxtime}");

            UpdateMaxMMDLenght();
        }

        /// <summary>
        /// 动作加载完毕调用的函数
        /// </summary>
        /// <param name="maxtime"></param>
        void OnMotionLoaded(ProgressController sender, float maxtime)
        {
            LogUtil.Debug($"Player::OnMotionLoaded:{maxtime}");

            UpdateMaxMMDLenght();
        }

        /// <summary>
        /// 更新MMD的最大长度
        /// </summary>
        void UpdateMaxMMDLenght()
        {
            if (_AudioPlayController != null)
            {
                LogUtil.Debug($"Player::UpdateMaxMMDLenght.MaxTime:{_AudioPlayController.MaxTime}");

                // 设置音频管理员的最大时间
                _AudioPlayController.MaxTime = GetMotionMaxTime(_AudioPlayController.MaxTime);

                // 如果没有设定结束时间，则将结束时间设定为最大动作时间
                if (CurrentItem?.AudioSetting?.EndTime <= 0)
                {
                    _AudioPlayController.endTime.val = GetMotionMaxTime();
                }

                LogUtil.Debug($"Player::CurrentItem?.AudioSetting?.EndTime2:{CurrentItem?.AudioSetting?.EndTime}");
                LogUtil.Debug($"Player::UpdateMaxMMDLenght.MaxTime2:{_AudioPlayController.MaxTime}");
                LogUtil.Debug($"Player::UpdateMaxMMDLenght.endTime2:{_AudioPlayController.endTime.val}");
            }
        }

        /// <summary>
        /// 获取人物原子列表
        /// </summary>
        public List<Atom> PersonAtoms
        {
            get
            {
                if (_PersonAtoms == null)
                {
                    _PersonAtoms = new List<Atom>();
                }

                return _PersonAtoms;
            }
        }

        internal Dictionary<int, PersonMotionController> PersonMotionControllers { get { return _PersonMotionControllers; } }
        internal CameraMotionController CameraMotionController { get { return _CameraMotionController; } }
        /// <summary>
        /// 获取或设置当前播放项目
        /// </summary>
        internal MMDEntity CurrentItem
        {
            get { return _CurrentItem; }
            set
            {
                if (_CurrentItem != value)
                {
                    _CurrentItem = value;
                }
            }
        }

        /// <summary>
        /// 当前项目数据内容改变的事件
        /// </summary>
        /// <param name="obj"></param>
        /// <exception cref="NotImplementedException"></exception>
        private void CurrentItem_Changed(MMDEntity obj)
        {
            LogUtil.Debug($"obj.InFavorite:{obj.InFavorite}");

            _FavoriteLabel.text = obj.InFavorite ? Lang.Get("UnFavorite") : Lang.Get("Favorite");
        }

        #region 配置相关的方法

        void SetIfChanged(JSONStorableStringChooser chooser, JSONNode json, string key)
        {
            if (chooser.val != chooser.defaultVal)
            {
                json[key] = chooser.val;
            }
            else if (json[key] != chooser.defaultVal)
            {
                json[key] = chooser.val;
            }
        }

        void SetIfChanged(JSONStorableBool chooser, JSONNode json, string key, bool not = false)
        {
            if (chooser.val != chooser.defaultVal)
            {
                if (not)
                {
                    json[key].AsBool = !chooser.val;
                }
                else
                {
                    json[key].AsBool = chooser.val;
                }
            }
        }

        void SetIfChanged(JSONStorableFloat chooser, JSONNode json, string key)
        {
            if (chooser.val != chooser.defaultVal)
            {
                json[key].AsFloat = chooser.val;
            }
        }

        private void FavoriteOrUnFavorite()
        {
            try
            {
                if (CurrentItem.InFavorite)
                {
                    CurrentItem.UnFavorite();
                }
                else
                {
                    CurrentItem.Favorite();
                }

                CurrentItem_Changed(CurrentItem);
            }
            catch (Exception ex)
            {
                LogUtil.LogError(ex, "Player:FavoriteOrUnFavorite::");
            }
        }

        /// <summary>
        /// 单个添加MMDEntity到播放列表
        /// </summary>
        /// <param name="item"></param>
        internal IEnumerator SaveSettings(MMDEntity item, bool dontSelect = false)
        {
            yield return null;

            if (item != null)
            {
                LogUtil.Debug($"Player::SaveSettingsToPlayList::Motions1:{item.Motions}");

                item = item.Clone();
                LogUtil.Debug($"Player::SaveSettingsToPlayList::Motions2:{item.Motions}");

                if (!dontSelect)
                {
                    SetValuesFromControllers(item);
                }

                if (isEditing)
                {
                    OutEditMode();
                }

                LogUtil.Debug($"AddPlayItem::::BEGIN");

                this.Playlist.AddPlayItem(item);
                //this.Playlist.Select(item);

                item.Save();

                // 保存后重新加载以激发更新事件，更新manager中的配置数据

                yield return null;

                //LoadMMD(item);
            }
        }

        /// <summary>
        /// 从Managers中更新设置值到对象中
        /// </summary>
        /// <param name="item"></param>
        private void SetValuesFromControllers(MMDEntity item)
        {
            // 从音频播放管理员获取音频和音频相关的配置，并将其更新到MMDEntity中
            SetIfChanged(_AudioPlayController.audioChooser, item.CurrentSetting.AudioSetting, "AudioPath");
            SetIfChanged(_AudioPlayController.startTime, item.CurrentSetting.AudioSetting, "StartTime");
            SetIfChanged(_AudioPlayController.endTime, item.CurrentSetting.AudioSetting, "EndTime");

            // 从镜头动作管理员获取镜头和镜头相关的配置，并将其更新到MMDEntity中
            SetIfChanged(_CameraMotionController.cameraChooser, item.CurrentSetting.CameraSetting, "CameraPath");
            SetIfChanged(_CameraMotionController.timeDelay, item.CurrentSetting.CameraSetting, "TimeDelay");

            SetIfChanged(_CameraMotionController.PositionOffsetXJSON, item.CurrentSetting.CameraSetting, "PositionOffsetX");
            SetIfChanged(_CameraMotionController.PositionOffsetYJSON, item.CurrentSetting.CameraSetting, "PositionOffsetY");
            SetIfChanged(_CameraMotionController.PositionOffsetZJSON, item.CurrentSetting.CameraSetting, "PositionOffsetZ");

            SetIfChanged(_CameraMotionController.RotationOffsetXJSON, item.CurrentSetting.CameraSetting, "RotationOffsetX");
            SetIfChanged(_CameraMotionController.RotationOffsetYJSON, item.CurrentSetting.CameraSetting, "RotationOffsetY");
            SetIfChanged(_CameraMotionController.RotationOffsetZJSON, item.CurrentSetting.CameraSetting, "RotationOffsetZ");

            // TODO 实现镜头跟随相关的数据记录

            // 从人物动作管理员获取人物动作及其配置数据，并将其更新到MMDEntity中
            for (int i = 0; i < _PersonAtoms.Count; i++)
            {
                var person = _PersonAtoms[i];
                var motionController = _PersonMotionControllers[person.GetInstanceID()];
                var motion = new PersonMotion();
                item.CurrentSetting.Motions[i] = motion;

                foreach (var motionChooser in motionController.MotionChoosers)
                {
                    if (motionChooser.val != motionChooser.defaultVal)
                    {
                        motion.Files.Add(motionChooser.val);
                    }
                }

                if (!config.LockPersonPosition)
                {
                    if (motionController.positionX.val != motionController.positionX.defaultVal)
                    {
                        motion.PositionX = motionController.positionX.val;
                    }
                    if (motionController.positionY.val != motionController.positionY.defaultVal)
                    {
                        motion.PositionY = motionController.positionY.val;
                    }
                    if (motionController.positionZ.val != motionController.positionZ.defaultVal)
                    {
                        motion.PositionZ = motionController.positionZ.val;
                    }

                    if (motionController.rotationX.val != motionController.rotationX.defaultVal)
                    {
                        motion.RotationX = motionController.rotationX.val;
                    }
                    if (motionController.rotationY.val != motionController.rotationY.defaultVal)
                    {
                        motion.RotationY = motionController.rotationY.val;
                    }
                    if (motionController.rotationZ.val != motionController.rotationZ.defaultVal)
                    {
                        motion.RotationZ = motionController.rotationZ.val;
                    }
                }

                SetIfChanged(motionController.enableFace, motion, "IgnoreFace", true);

                SetIfChanged(motionController.motionScale, motion, "MotionScale");

                SetIfChanged(motionController.UseAllJointsSettingsJSON, motion, "UseAllJointsSettings");
                SetIfChanged(motionController.AllJointsSpringPercentJSON, motion, "AllJointsSpringPercent");
                SetIfChanged(motionController.AllJointsDamperPercentJSON, motion, "AllJointsDamperPercent");
                SetIfChanged(motionController.AllJointsMaxVelocityJSON, motion, "AllJointsMaxVelocity");

                SetIfChanged(motionController.timeDelay, motion, "TimeDelay");
            }
        }

        /// <summary>
        /// 清空播放列表
        /// </summary>
        void ClearPlayList()
        {
            try
            {
                StopPlaying();
                CurrentItem = null;

                LogUtil.Debug($"Player::Clear:ClearPlayList.");
                ClearMotions();

                _PlaylistChooser.displayChoices = noneStrings;
                _PlaylistChooser.choices = noneStrings;
                _PlaylistChooser.valNoCallback = noneString;

                ShowInitLoadUI();
            }
            catch (Exception ex)
            {
                LogUtil.Debug(ex, $"ClearList Error");
            }
        }

        /// <summary>
        /// 清空选中的元素
        /// </summary>
        private void ClearMotions()
        {
            _AudioPlayController.Clear();
            _CameraMotionController.Clear();

            foreach (var item in _PersonMotionControllers)
            {
                item.Value.Reset();
            }
        }

        #endregion

        #region 播放相关的方法

        public new void Start()
        {
            LogUtil.Debug("Player::START!!!");

            if (containingAtom.type == "Person")
            {
                _CameraMotionController.FocusAtomJSON.val = containingAtom.uid;
                _CameraMotionController.FocusReceiverJSON.val = "neckControl";
            }

            _DebugUIs.RefreshView();
            ShowInitLoadUI();

            HideStatusMessage();

            // 从默认播放列表加载内容
            StartCoroutine(this.Playlist.LoadFromDefalut());
        }

        public void Update()
        {
            if (IsLocked) return;

            // 如果按了SPACE键则切换播放状态
            if (Input.GetKeyDown(_PlayToggleKey) && config.EnableSpacePlay)
            {
                TogglePlaying();
                return;
            }
            // 左箭头后退1s
            if (Input.GetKeyUp(KeyCode.LeftArrow) && config.EnableLeftArrow)
            {
                _AudioPlayController.Back();
                return;
            }
            // 右箭头前进1s
            if (Input.GetKeyUp(KeyCode.RightArrow) && config.EnableRightArrow)
            {
                _AudioPlayController.Forward();
                return;
            }

            // 下箭头 下一个
            if (Input.GetKeyUp(KeyCode.UpArrow) && config.EnableUpArrow)
            {
                this.Prev();
                return;
            }

            // 下箭头 下一个
            if (Input.GetKeyUp(KeyCode.DownArrow) && config.EnableDownArrow)
            {
                this.Next();
                return;
            }

            // 如果选中显示Debug信息
            if (_ShowDebugInfoJSON.val)
            {
                var playInfo = DateTime.Now.ToString();

                playInfo += $"\r\n" +
                    $"_IsPlaying:{_IsPlaying.val}" +
                    $"\r\n" +
                    $"_IsLoading:{_IsLoading}" +
                    $"\r\n" +
                    $"Playing:{_AudioPlayController.IsPlaying}" +
                    $"\r\n" +
                    $"Loading:{_AudioPlayController.IsLoading}" +
                    $"\r\n" +
                    $"Waiting:{_AudioPlayController.IsWaitingPlay}" +
                    $"\r\n" +
                    $"IsEnd:{_AudioPlayController.IsEnd}" +
                    $"\r\n" +
                    $"IsPaused:{_AudioPlayController.IsPaused.val}" +
                    $"\r\n" +
                    $"HasAudio:{_AudioPlayController.HasAudio}" +
                    $"\r\n" +
                    $"Clip:{_AudioPlayController.PlayingAudio?.displayName}" +
                    $"\r\n" +
                    $"CheckTime:{_CheckTime}" +
                    $"\r\n" +
                    $"MaxTime:{_AudioPlayController.MaxTime}" +
                    $"\r\n" +
                    $"Progress:{_AudioPlayController.Progress}" +
                    $"\r\n" +
                    $"endTime:{_AudioPlayController.endTime.val}" +
                    $"\r\n" +
                    $"AMaxTime:{_AudioPlayController.MaxTime}" +
                    $"\r\n" +
                    $"CMaxTime:{_CameraMotionController.MaxTime}" +
                    $"\r\n";

                foreach (var m in _PersonMotionControllers)
                {
                    playInfo +=
                        $"PMaxTime:{m.Value.MaxTime}" +
                        $"\r\n";
                }

                _DebugInfo.SetVal(playInfo);
            }

            // 如果冻结动画或不是活动和启用状态，停止播放后，直接返回
            if (SuperController.singleton.freezeAnimation || !SuperController.singleton.isActiveAndEnabled || _PlaySpeedJSON.val <= 0f)
            {
                _CheckFreezeTime += Time.deltaTime;
                _AudioPlayController.AudioPause();
                return;
            }

            // 如果有冻结时间，则说明已经暂停，此时取消暂停
            if (_CheckFreezeTime > 0 && !isPaused)
            {
                _AudioPlayController.AudioUnPause();
                _CheckFreezeTime = 0f;

                return;
            }

            // 如果音频还在加载中
            if (_AudioPlayController.IsLoading)
            {
                _CheckTime += Time.deltaTime;
                if (_CheckTime > MAX_CHECK_TIME)
                {
                    _CheckTime = 0f;
                    _AudioPlayController.CheckLoad();
                }
                return;
            }

            // 检查是否可以播放
            if (!CanPlay()) return;

            try
            {
                _AudioPlayController.Update(Time.deltaTime, _PlaySpeedJSON.val);
            }
            catch (Exception ex)
            {
                LogUtil.Debug(ex, "Update:");
                this.StopPlaying();
            }
        }

        public void FixedUpdate()
        {
        }

        /// <summary>
        /// 获取动作最大时间
        /// </summary>
        /// <returns></returns>
        float GetMotionMaxTime(float maxTime = 0)
        {
            var cameraMaxTime = _CameraMotionController.MaxTime;
            var motionMaxTime = _PersonMotionControllers.Values.Max(m => m.MaxTime);
            return Mathf.Max(cameraMaxTime, motionMaxTime, maxTime);
        }

        void Prev()
        {
            if (_PersonMotionControllers.Values.Any(m => !m.IsReady))
            { return; }

            this.Playlist.Prev();
        }

        /// <summary>
        /// 下一个
        /// </summary>
        void Next()
        {
            if (_PersonMotionControllers.Values.Any(m => !m.IsReady))
            { return; }

            LogUtil.Debug($"Player::NEXT!");

            if (isEditing)
            {
                StartCoroutine(Repeat());
            }
            else
            {
                this.Playlist.Next();
            }
        }

        /// <summary>
        /// 设置镜头动作进度
        /// </summary>
        /// <param name="value"></param>
        private void SetMMDCameraProgress(float value)
        {
            if (config.EnableCamera)
            {
                //var fixedview = false;
                if (_CameraMotionController.cameraChooser.val == noneString)
                {
                    //fixedview = true;

                    //// 启用镜头跟随
                    //AlignHeadToCamera(fixedview);
                }
                else if (_CameraMotionController != null)
                {
                    // 启用镜头跟随
                    AlignHeadToCamera();
                    _CameraMotionController.SetProgress(value);

                    return;
                }
            }

            _CameraMotionController.IsActive = false;
        }

        /// <summary>
        /// 判断当前状态是否可以播放
        /// </summary>
        /// <returns></returns>
        private bool CanPlay()
        {
            return (_IsPlaying.val) && !_AudioPlayController.IsLoading && !_AudioPlayController.IsEnd && !_AudioPlayController.IsPaused.val && !_IsLoading;
        }

        /// <summary>
        /// 重新播放
        /// </summary>
        private IEnumerator Repeat()
        {
            yield return null;
            //SetAllPersonOff();
            yield return null;
            StopAndStartOver();
            yield return new WaitForSeconds(1);
            SetAllPersonOn();
            yield return null;
            _AudioPlayController.Play();
        }

        /// <summary>
        /// 重置动作
        /// </summary>
        /// <param name="atom"></param>
        private void ResetPose(Atom atom)
        {
            atom.ResetRigidbodies();
            atom.ResetPhysical();

            AtomUI componentInChildren = atom.GetComponentInChildren<AtomUI>();
            if (componentInChildren != null)
            {
                componentInChildren?.resetButton?.onClick?.Invoke();
            }
        }

        bool isEditing = false;

        bool isPaused = true;

        /// <summary>
        /// 切换播放和暂停状态
        /// </summary>
        public bool TogglePlaying()
        {
            // 如果正在播放
            if (_IsPlaying.val)
            {
                // 如果已经暂停
                if (_AudioPlayController.IsPaused.val)
                {
                    _AudioPlayController.AudioUnPause();
                    isPaused = false;
                }
                else
                {
                    this.Pause();
                }

            }
            else // 如果未在播放
            {
                // 开始播放
                StartPlaying();
            }

            // 返回是否在播放状态
            bool onPlay = SetPlayButton();

            return onPlay;
        }

        /// <summary>
        /// 暂停播放
        /// </summary>
        public void Pause()
        {
            if (_IsPlaying.val)
            {
                // 暂停播放
                _AudioPlayController.AudioPause();
                isPaused = true;
            }
        }

        private bool MainHUDVisible
        {
            get
            {
                //#if (VAM_GT_1_20)
                return (config.DeactiveCameraWhenMainHUDOpened && SuperController.singleton.mainHUD.gameObject.activeSelf);
                //#else
                //                return (config.DeactiveCameraWhenMainHUDOpened && SuperController.singleton.MainHUDVisible);
                //#endif
            }
        }

        private bool SetPlayButton()
        {
            var onPlaying = _IsPlaying.val && !_AudioPlayController.IsPaused.val;

            // 如果是播放状态，切换播放按钮的文字
            if (onPlaying)
            {
                if (_UIPlayButton.buttonText.text != Lang.Get("Pause"))
                {
                    _UIPlayButton.buttonText.text = Lang.Get("Pause");

                    if (!CameraActive)
                    {
                        _CameraMotionController.DisableNavigation(false);
                    }
                    else
                    {
                        _CameraMotionController.DisableNavigation();
                    }
                }
            }
            else
            {
                if (_UIPlayButton.buttonText.text != Lang.Get("Play"))
                {
                    _UIPlayButton.buttonText.text = Lang.Get("Play");
                    _CameraMotionController.DisableNavigation(false);
                }
            }
            return onPlaying;
        }

        /// <summary>
        /// 停止播放
        /// </summary>
        public void StopPlaying()
        {
            _IsPlaying.val = false;
            _AudioPlayController.Stop(4);
        }

        /// <summary>
        /// 停止播放并定位到开头
        /// </summary>
        public void StopAndStartOver()
        {
            _AudioPlayController.Stop(5);

            // 重置进度到开头
            _AudioPlayController.StartAudioOver();
        }

        /// <summary>
        /// 开始播放
        /// </summary>
        public void StartPlaying()
        {
            isPaused = false;
            // 如果正在播放 直接返回
            if (_IsPlaying.val)
            {
                if (_AudioPlayController.IsPaused.val)
                {
                    _AudioPlayController.AudioUnPause();
                }

                return;
            }

            _IsPlaying.val = true;
        }

        void LoadMMD(MMDEntity mmdItem)
        {
            LogUtil.Debug($"Player::LoadMMD:{mmdItem?.Key},SettingIndex:{mmdItem?.SettingIndex}");

            if (mmdItem == null)
            {
                this.ClearPlayList();
            }
            else
            {
                this.LoadMMD(mmdItem, false);
            }
        }
        /// <summary>
        /// 加载MMD数据
        /// </summary>
        /// <param name="entity"></param>
        internal void LoadMMD(MMDEntity entity, bool test)
        {
            try
            {
                ClearMotions();
                if (entity == null)
                {
                    // TODO 传入空对象时，清空配置

                    //ClearPlayList();
                }
                else
                {
                    NotifyMMDSelected(entity);

                    if (!test)
                    {
                        this.ShowPlayUI();
                    }
                    else
                    {
                        this.ShowEditUI();
                    }
                }
            }
            catch (Exception e)
            {
                LogUtil.LogError(e, "Player::LoadMMD");
            }
        }

        /// <summary>
        /// 播放MMD数据
        /// </summary>
        /// <param name="entity"></param>
        internal IEnumerator PlayMMD(MMDEntity entity)
        {
            if (entity == null)
            {
                yield break;
            }

            CurrentItemChangedNoConfirm?.Invoke(entity, null);

            if (CurrentItem != entity)
            {
                CurrentItem = entity;

                CurrentItem_Changed(entity);
            }

            LogUtil.Debug($"Player::PlayMMD:{entity.Title}");

            var fileData = entity.GetFileData();

            // 如果没有内容，停止播放
            if (fileData.DefaultCamera == noneString
                && fileData.DefaultAudio == noneString
                && fileData.DefaultMotion == noneString)
            {
                this.StopPlaying();
            }

            // 加载镜头动作
            LoadCameraSettings(entity.CameraSetting, fileData.MotionPaths, fileData.MotionNames);

            LogUtil.Debug($"LoadMMD::PersonAtoms:{PersonAtoms.Count}");

            for (var i = 0; i < PersonAtoms.Count; i++)
            {
                LogUtil.Debug($"LoadMMD::Motions:{entity.Motions.Count}");

                // 如果没有足够的动作，则跳过后边的人物设定
                if (entity.Motions.Count <= i)
                {
                    entity.Motions.Add(new PersonMotion());
                }

                var atom = PersonAtoms[i];

                LogUtil.Debug($"LoadMMD::PersonAtom:{atom.uid}");

                var motion = entity.Motions[i] as PersonMotion;

                LogUtil.Debug($"LoadMMD::motion:{motion.Files.Count}");

                // 设置人物动作
                yield return StartCoroutine(LoadPersonMotion(atom, motion, fileData.MotionPaths, fileData.MotionNames));

            }

            // 加载音频设置
            LoadAudioSettings(entity.AudioSetting, fileData.AudioPaths, fileData.AudioNames);

            yield return null;//new WaitForSeconds(1);
            _AudioPlayController.Forward(0.001f);
            yield return new WaitForSeconds(1);

            foreach (var atom in PersonAtoms)
            {
                SetPersonOff(atom);
            }

            yield return new WaitForSeconds(1);
            //_AudioPlayController.SetProgress(0.0f);

            foreach (var atom in PersonAtoms)
            {
                SetPersonOn(atom);
            }

            _IsLoading = false;

            if (_WaitingForPlay)
            {
                _WaitingForPlay = false;

                this.StartPlaying();
            }
        }

        /// <summary>
        /// 加载音频设置
        /// </summary>
        /// <param name="settings"></param>
        /// <param name="choices"></param>
        /// <param name="displayChoices"></param>
        private void LoadAudioSettings(AudioSetting settings, List<string> choices, List<string> displayChoices)
        {
            LogUtil.Debug($"Player::AudioSettings - StartTime:{settings.StartTime},EndTime:{settings.EndTime}");

            var choice = string.IsNullOrEmpty(settings.AudioPath) ? noneString : settings.AudioPath;

            _AudioPlayController.InitPlay(choices, displayChoices, choice, settings.StartTime, settings.EndTime);
        }

        /// <summary>
        /// 加载镜头设置
        /// </summary>
        /// <param name="settings"></param>
        /// <param name="choices"></param>
        /// <param name="displayChoices"></param>
        private void LoadCameraSettings(CameraSetting settings, List<string> choices, List<string> displayChoices)
        {
            var choice = string.IsNullOrEmpty(settings.CameraPath) ? noneString : settings.CameraPath;

            _CameraMotionController.InitPlay(choices, displayChoices, choice, settings);
        }

        /// <summary>
        /// 设置人物动画
        /// </summary>
        /// <param name="person"></param>
        /// <param name="motion"></param>
        private IEnumerator LoadPersonMotion(Atom person, PersonMotion motion, List<string> motionList, List<string> motionNameList)
        {
            // 实例化MMD人物
            var personMotionController = InitPersonMotionController(person);

            personMotionController.InitSettings(motionList, motionNameList, motion);

            yield return StartCoroutine(SetMMDProgress(personMotionController, 0));

            //yield return null;

            //SetPersonOff(person);

            //yield return new WaitForSeconds(1f);

            //SetPersonOn(person);
        }

        /// <summary>
        /// 初始化人物动作管理器
        /// </summary>
        /// <param name="atom"></param>
        /// <param name="atomid"></param>
        /// <returns></returns>
        private PersonMotionController InitPersonMotionController(Atom atom)
        {
            try
            {
                if (atom?.type == "Person")
                {
                    if (!PersonAtoms.Contains(atom))
                    {
                        // 将人物原子添加到人物原子列表中
                        PersonAtoms.Add(atom);
                    }

                    var atomid = atom.GetInstanceID();

                    if (_PersonMotionControllers.ContainsKey(atomid))
                    {
                        return _PersonMotionControllers[atomid];
                    }
                    else
                    {
                        PersonMotionController controller = new PersonMotionController(atom, _PluginPath + "/g2f.pmx");
                        controller.player = this;

                        // 将MMD人物添加到MMD人物字典
                        _PersonMotionControllers[atomid] = controller;

                        CreatePersonMotionUI(controller, atom);

                        controller.OnLoaded += OnMotionLoaded;

                        LoadMMD(CurrentItem);

                        return controller;
                    }
                }
            }
            catch (Exception ex)
            {
#if DEBUG
                LogUtil.LogError(ex, "Player::InitPersonMotionController:");
#endif
            }

            return null;
        }

        internal void ResetPersonControllerState()
        {
            foreach (var personMotionController in PersonMotionControllers)
            {
                personMotionController.Value.ResetControlState();
            }
        }

        /// <summary>
        /// 移除惹怒我动作控制器
        /// </summary>
        /// <param name="atom"></param>
        private void RemovePersonMotionController(Atom atom)
        {
            try
            {
                if (atom?.type == "Person")
                {
                    if (PersonAtoms.Contains(atom))
                    {
                        PersonAtoms.Remove(atom);
                    }
                    var atomid = atom.GetInstanceID();

                    PersonMotionController controller;

                    if (_PersonMotionControllers.TryGetValue(atomid, out controller))
                    {
                        controller.OnLoaded -= OnMotionLoaded;

                        RemovePersonMotionUI(controller, atom);

                        controller.OnDestroy();

                        _PersonMotionControllers.Remove(atomid);
                    }
                }
            }
            catch (Exception ex)
            {
#if DEBUG
                LogUtil.LogError(ex, $"Player::RemovePersonMotionController");
#endif
            }
        }

        /// <summary>
        /// 打开人物
        /// </summary>
        /// <param name="person"></param>
        private void SetPersonOn(Atom person)
        {
            // 碰撞开启
            person.collisionEnabled = true;

            PersonMotionController controller;

            if (_PersonMotionControllers.TryGetValue(person.GetInstanceID(), out controller))
            {
                StartCoroutine(controller.Ready());
            }
        }

        /// <summary>
        /// 关闭人物
        /// </summary>
        /// <param name="person"></param>
        private void SetPersonOff(Atom person)
        {
            if (config.ResetPhysicalWhenLoadMotion)
            {
                // 重置物理
                person.ResetRigidbodies();
                person.ResetPhysical();
            }

            // 关闭碰撞
            person.collisionEnabled = false;

            PersonMotionController controller;

            if (_PersonMotionControllers.TryGetValue(person.GetInstanceID(), out controller))
            {
                StartCoroutine(controller.MakeReady());
            }
        }

        /// <summary>
        /// 重置所有人物动作
        /// </summary>
        /// <returns></returns>
        IEnumerator ResetAllPersonMotion()
        {
            foreach (var item in PersonMotionControllers)
            {
                ResetPose(item.Value.PersonAtom);

                yield return null;

                item.Value.ReloadMotions(init: true);

                yield return new WaitForSeconds(1);

                SetPersonOff(item.Value.PersonAtom);

                yield return new WaitForSeconds(1);

                SetPersonOn(item.Value.PersonAtom);
            }
        }

        /// <summary>
        /// 设置所有人物关闭
        /// </summary>
        private void SetAllPersonOff()
        {
            foreach (var atom in PersonAtoms)
            {
                SetPersonOff(atom);
            }
        }

        /// <summary>
        /// 设置所有人物开启
        /// </summary>
        private void SetAllPersonOn()
        {
            foreach (var atom in PersonAtoms)
            {
                SetPersonOn(atom);
            }
        }

        /// <summary>
        /// 设置动作播放进度
        /// </summary>
        /// <param name="atomid"></param>
        /// <param name="value"></param>
        /// <returns></returns>
        private IEnumerator SetMMDProgress(PersonMotionController mmdPerson, float value)
        {
            yield return null;

            try
            {
                mmdPerson.SetProgress(value);
            }
            catch (Exception e)
            {
                LogUtil.Debug(e, $"SetMMDProgress:");
            }
        }

        #endregion

        #region 镜头播放的代码

        float randomTick = 0;
        float randomTimeMax = 3f;
        float randomTimeMin = 1f;

        Vector3 randomPosition = new Vector3(0f, 1f, 3f);
        float randomFov = 40f;

        string[] randomTargets = new string[] { "neckControl", "hipControl", "pelvisControl", "chestControl", "headControl", "abdomenControl", "abdomen2Control" };

        string tagetId = "hipControl";

        internal bool CameraActive
        {
            get
            {
                return !(config.DeactiveCameraWhenMainHUDOpened && MainHUDVisible);
            }
        }

        public void AlignHeadToCamera(bool fixedview = false)
        {
            try
            {
                // 如果主HUD激活，则停止跟踪（测试时不启用）
                if (!CameraActive)
                {
                    _CameraMotionController.IsActive = false;
                }
                else
                {
                    _CameraMotionController.IsActive = true;

                    if (fixedview)
                    {
                        randomTick += Time.deltaTime;

                        if (randomTick > UnityEngine.Random.Range(randomTimeMin, randomTimeMax))
                        {
                            randomTick = 0f;
                            _CameraMotionController.Update(randomPosition, randomFov, Quaternion.identity, tagetId);
                        }
                    }
                }
            }
            catch (Exception e)
            {
                LogUtil.Debug(e, "Failed to update: ");
            }
        }

        #endregion

        /// <summary>
        /// 当启用时执行的方法
        /// </summary>
        public override void OnEnable()
        {
            LogUtil.Debug("---------------OnEnable!!!!!");

            base.OnEnable();

            _AudioPlayController?.OnEnable();
            _CameraMotionController?.OnEnable();
            _MMDFolderController?.OnEnable();
            if (_PersonMotionControllers != null)
            {
                foreach (var m in _PersonMotionControllers)
                {
                    m.Value.OnEnable();
                }
            }
        }

        /// <summary>
        /// 当禁用时执行的方法
        /// </summary>
        public override void OnDisable()
        {
            LogUtil.Debug("---------------OnDisable!!!!!");

            base.OnDisable();

            _AudioPlayController?.OnDisable();
            _CameraMotionController?.OnDisable();
            _MMDFolderController?.OnDisable();

            if (_PersonMotionControllers != null)
            {
                foreach (var m in _PersonMotionControllers)
                {
                    m.Value.OnDisable();
                }

                _PersonMotionControllers.Clear();
            }
        }

        /// <summary>
        /// 当销毁时执行的方法
        /// </summary>
        public override void OnDestroy()
        {
            LogUtil.Debug("---------------OnDestroy!!!!!");

            base.OnDestroy();

            DestroyStatusUI();

            RemovePlaylistEvent();

            _AudioPlayController.OnLoaded -= OnAudioLoaded;

            _AudioPlayController.OnDestroy();
            _AudioPlayController = null;
            _CameraMotionController.OnLoaded -= OnCameraLoaded;

            _CameraMotionController.OnDestroy();
            _CameraMotionController = null;
            foreach (var m in _PersonMotionControllers)
            {
                m.Value.OnLoaded += OnMotionLoaded;

                m.Value.OnDestroy();
            }
            _MMDFolderController.OnMMDImported -= OnMMDImported;

            _MMDFolderController.OnDestroy();
            _MMDFolderController = null;

            this.CurrentItem = null;

            this._PersonMotionControllers.Clear();
            this._PersonMotionControllers = null;

            this._PersonAtoms.Clear();
            this._PersonAtoms = null;

            this._PlayUIs.Clear();
            this._PlayUIs = null;

            this._UIPlayButton = null;

            SuperController.singleton.BroadcastMessage("OnActionsProviderDestroyed", this, SendMessageOptions.DontRequireReceiver);
        }
    }
}
